#!/usr/bin/env python3
# thoth-s2i
# Copyright(C) 2020 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Helpers used in thoth-s2i library."""

from typing import Any
from typing import List
import logging
import subprocess

_LOGGER = logging.getLogger(__name__)


def _subprocess_run(args: List[str], **kwargs: Any) -> Any:
    """Run the given command as a subprocess - a thin wrapper."""
    _LOGGER.debug("Executing command %r with subprocess flags %r", args, kwargs)
    return subprocess.run(args, universal_newlines=True, **kwargs)



#!/usr/bin/env python3
# thoth-s2i
# Copyright(C) 2020 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Convert OpenShift Python s2i buildconfigs to use Thoth's s2i."""

from pathlib import Path
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
import functools
import logging
import os
import re
import requests
import subprocess
import sys

from .exceptions import ImportImageError
from .exceptions import OCError
from .exceptions import S2I2ThothException
from .objs import ImageStream
from .objs import BuildConfig
from .helpers import _subprocess_run

_LOGGER = logging.getLogger(__name__)

_OC_CHECKED = False
_THOTH_S2I_README = os.getenv(
    "THOTH_S2I_README", "https://raw.githubusercontent.com/thoth-station/s2i-thoth/master/README.rst",
)
_RE_S2I = re.compile(r"quay.io/thoth-station/s2i-thoth-\S*")


def oc_check() -> None:
    """Check if oc (OpenShift client binary) is available and if the user is logged in into the cluster."""
    try:
        oc_version = _subprocess_run(["oc", "version"], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    except FileNotFoundError as exc:
        raise OCError(
            f"Failed to obtain information about OpenShift client - make sure oc is "
            f"installed and available on PATH: {str(exc)}"
        ) from exc

    if oc_version.returncode != 0:
        raise OCError(f"Failed to obtain information about OpenShift client: {oc_version.stderr}")


@functools.wraps(oc_check)
def oc_check_once(func: Callable[..., Any]) -> Callable[..., Any]:
    """Check oc binary for required functions once this library is run."""
    def inner(*args: Any, **kwargs: Any) -> Any:
        global _OC_CHECKED

        if not _OC_CHECKED:
            oc_check()
            _OC_CHECKED = True

        return func(*args, **kwargs)

    return inner


def _do_import_image(namespace: str, image: str, stdout: Any = None) -> None:
    """Import an image to namespace."""
    parts = image.rsplit("/", maxsplit=1)
    parts = parts[-1].rsplit(":", maxsplit=1)
    imagestream_name = parts[0]

    subcommand = _subprocess_run(
        ["oc", "import-image", "--namespace", namespace, imagestream_name, "--from", image, "--confirm"],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    if subcommand.returncode != 0:
        raise ImportImageError(f"Failed to import image {image}: {subcommand.stderr}")

    if stdout is not None:
        print(subcommand.stdout, file=stdout)


@oc_check_once
def import_thoth_s2i_image(namespace: str, image: Optional[str] = None, stdout: Any = sys.stdout) -> None:
    """Import Thoth's s2i image into the given namespace."""
    if image is not None:
        _do_import_image(namespace, image, stdout=stdout)
        return

    for image in get_thoth_s2i_images():
        _do_import_image(namespace, image, stdout=stdout)


def get_thoth_s2i_images() -> List[str]:
    """Get a listing of Thoth's s2i images available."""
    response = requests.get(_THOTH_S2I_README)
    try:
        response.raise_for_status()
    except Exception as exc:
        raise S2I2ThothException("Failed to obtain Thoth's s2i images from GitHub") from exc

    return sorted(set(_RE_S2I.findall(response.text)), reverse=True)


@oc_check_once
def oc_get_bc(namespace: str, selector: Optional[str] = None, path: str = "buildconfigs.yaml") -> None:
    """Get buildconfigs installed in the given namespace to the given file described by path."""
    if not os.path.isfile(path):
        try:
            Path(path).touch()
        except Exception as exc:
            raise S2I2ThothException(f"Cannot create file {path} for storing buildconfigs") from exc

    cmd = ["oc", "get", "bc", "--namespace", namespace, "-o", "yaml"]

    if selector:
        cmd.extend(("-l", selector))

    with open(path, "w") as output_file:
        subcommand = _subprocess_run(cmd, stdout=output_file.fileno(), stderr=subprocess.PIPE)

    if subcommand.returncode != 0:
        raise OCError(f"Failed to obtain buildconfigs from namespace {namespace}: {subcommand.stderr}")


@oc_check_once
def oc_get_is(namespace: str, selector: Optional[str] = None, path: str = "imagestreams.yaml") -> None:
    """Get ImageStreams installed in the given namespace to the given file described by path."""
    if not os.path.isfile(path):
        try:
            Path(path).touch()
        except Exception as exc:
            raise S2I2ThothException(f"Cannot create file {path} for storing imagestreams") from exc

    cmd = ["oc", "get", "is", "--namespace", namespace, "-o", "yaml"]

    if selector:
        cmd.extend(("-l", selector))

    with open(path, "w") as output_file:
        subcommand = _subprocess_run(cmd, stdout=output_file.fileno(), stderr=subprocess.PIPE)

    if subcommand.returncode != 0:
        raise OCError(f"Failed to obtain imagestreams from namespace {namespace}: {subcommand.stderr}")


def get_build_usage_report(
    image_streams: Dict[str, ImageStream], build_configs: Dict[str, BuildConfig]
) -> Dict[str, Any]:
    """Get report usage of build configs."""
    report_dict = {}
    for build_config in build_configs.values():
        strategy_type = build_config.raw.get("spec", {}).get("strategy", {}).get("type")

        report_dict[build_config.name] = {
            "strategy": strategy_type,
            "image_stream": None,
            "image_stream_tag": None,
            "is_s2i": strategy_type.lower() == "source",
            "is_s2i_thoth": None,
            "s2i_image_tag_imported": None,
            "s2i_image_name": None,
            "s2i_image_tag": None,
        }

        if not report_dict[build_config.name]["is_s2i"]:
            _LOGGER.debug(
                "Skipping s2i detection for %r as build strategy is not s2i", build_config.name,
            )
            continue

        source_strategy = build_config.get_source_strategy()
        kind = source_strategy.get("from", {}).get("kind")
        if not kind:
            _LOGGER.error("Cannot determine kind for source strategy for %r", build_config.name)
            continue

        if kind != "ImageStreamTag":
            continue

        is_name = source_strategy.get("from", {}).get("name")
        is_name, is_tag = is_name.rsplit(":", maxsplit=1)

        report_dict[build_config.name]["image_stream"] = is_name
        report_dict[build_config.name]["image_stream_tag"] = is_tag

        image_stream = image_streams.get(is_name)
        if not image_stream:
            _LOGGER.error(
                "Cannot find image stream %r used by build config %r", is_name, build_config.name,
            )
            continue

        for tag_record in image_stream.raw.get("spec", {}).get("tags", []):
            if tag_record.get("name") == is_tag:
                image_record = tag_record.get("from", {}).get("name")
                if image_record:
                    parts = image_record.rsplit(":", maxsplit=1)
                    if len(parts) == 2:
                        image, tag = parts
                    else:
                        image, tag = image_record, "latest"
                    report_dict[build_config.name]["s2i_image_tag"] = tag
                    report_dict[build_config.name]["s2i_image_name"] = image
                else:
                    _LOGGER.error(
                        "Cannot parse image record for image stream %r tag %r", is_name, is_tag,
                    )
                break
        else:
            _LOGGER.error(
                "Cannot find tag %r in image stream %r used by %r", is_tag, is_name, build_config.name,
            )
            continue

    s2i_thoth = get_thoth_s2i_images()
    for entry in report_dict.values():
        if not entry["is_s2i"]:
            continue

        if entry["s2i_image_name"] in s2i_thoth:
            entry["is_s2i_thoth"] = True

    return report_dict


def change_image_stream(
    build_configs: Dict[str, BuildConfig], new: str, cond_func: Callable[[BuildConfig], bool],
) -> Dict[str, BuildConfig]:
    """Change image stream in BuildConfigs based on condition."""
    result = {}
    for build_config in build_configs.values():
        if not cond_func(build_config):
            # XXX: do we want something more reliable?!
            continue

        _LOGGER.info(
            "Patching BuildConfig %r, replacing %r with %r",
            build_config.name,
            build_config.get_image_stream_tag(),
            new,
        )
        build_config.replace_in_file(build_config.get_image_stream_tag(), new)
        build_config.set_image_stream_tag(new)
        result[build_config.name] = build_config

    return result



#!/usr/bin/env python3
# thoth-s2i
# Copyright(C) 2020 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Routines needed for parsing and representing OpenShift's BuildConfig."""

import os
import logging
import yaml
import subprocess
from typing import Any
from typing import Dict
from typing import Mapping
from typing import Generator
from typing import List
from typing import Optional
from typing import Union

import attr

from .exceptions import S2I2ThothException
from .exceptions import OCError
from .helpers import _subprocess_run

_LOGGER = logging.getLogger(__name__)

# A YAML safe loader that does not parse datetime.
_NoDatesSafeLoader = yaml.SafeLoader
_NoDatesSafeLoader.yaml_implicit_resolvers = {
    k: [r for r in v if r[0] != "tag:yaml.org,2002:timestamp"]
    for k, v in _NoDatesSafeLoader.yaml_implicit_resolvers.items()
}


@attr.s(slots=True)
class OpenShiftObject:
    """A wrapper for BuildConfig representation."""

    file_path = attr.ib(type=str)
    raw = attr.ib(type=Dict[str, Any])

    @classmethod
    def load_file(cls, path: str, skip_errors: bool = True) -> Dict[str, Any]:
        """Load the given OpenShift object from a file."""
        raise NotImplementedError

    @classmethod
    def load_all(cls, path: str, skip_errors: bool = True) -> Mapping[str, "OpenShiftObject"]:
        """Load the given OpenShift object from a file or files present in a directory, recursively."""
        if os.path.isfile(path):
            return cls.load_file(path)

        if not os.path.isdir(path):
            raise S2I2ThothException(f"Path {path!r} is not a file or directory")

        result: Dict[str, "OpenShiftObject"] = {}
        for root, _, files in os.walk(path, followlinks=True):
            for file in files:
                instances = cls.load_file(os.path.join(root, file), skip_errors=skip_errors)
                for instance in instances.values():
                    if instance.name in result:
                        if not skip_errors:
                            raise S2I2ThothException(f"Multiple definitions of {instance.name!r} found")

                        _LOGGER.warning(
                            f"Multiple definitions of %r found, skipping...", instance.name,
                        )
                        continue

                    result[instance.name] = instance

        return result

    def labels(self) -> Dict[str, str]:
        """Get labels applied to build config."""
        result: Dict[str, str] = self.raw.get("metadata", {}).get("labels", {})
        return result

    @property
    def name(self) -> str:
        """Get name of this object."""
        result: str = self.raw["metadata"]["name"]
        return result

    @staticmethod
    def _load_file_content(
        path: str, skip_errors: bool = True
    ) -> Optional[Union[Dict[str, Any], List[Dict[str, Any]]]]:
        """Load content of the given file."""
        _LOGGER.info("Loading file content from %r", path)

        content = None
        try:
            with open(path, "r") as input_file:
                content = yaml.load(input_file, Loader=_NoDatesSafeLoader)  # use BaseLoader not to parse datetime.
        except Exception as exc:
            if not skip_errors:
                raise

            _LOGGER.warning(
                "Failed to load and/or parse file %r, the file will be skipped: %s:", path, str(exc),
            )

        if content is None:
            if not skip_errors:
                raise S2I2ThothException(f"File {path!r} has value null")

            _LOGGER.warning("File %r holds null, skipping...", path)
            return None

        if not isinstance(content, (list, dict)):
            if not skip_errors:
                raise S2I2ThothException(
                    f"File {path!r} holds a string, does not look like OpenShift YAML/JSON template"
                )

            _LOGGER.warning(
                "File %r holds a string, does not look like OpenShift YAML/JSON template", path,
            )
            return None

        return content

    @classmethod
    def _iter_objects(
        cls, content: Union[Dict[str, Any], List[Dict[str, Any]]]
    ) -> Generator[Dict[str, Any], None, None]:
        """Iterate over objects found in the template."""
        if isinstance(content, list):
            for item in content:
                if isinstance(item, (dict, list)):
                    yield from cls._iter_objects(item)
            return

        if content.get("kind", "").strip().lower() == "template":
            for obj in content.get("objects", []):
                yield from cls._iter_objects(obj)
            return

        if content.get("kind", "").strip().lower() == "list":
            for obj in content.get("items", []):
                yield from cls._iter_objects(obj)
            return

        yield content

    @classmethod
    def _instantiate_objects(
        cls,
        path: str,
        content: Union[Dict[str, Any], List[Dict[str, Any]]],
        class_: type,
        kind: str,
        skip_errors: bool = True,
    ) -> Mapping[str, "OpenShiftObject"]:
        """Instantiate objects from from loaded content."""
        result: Dict[str, "OpenShiftObject"] = {}
        for obj in cls._iter_objects(content):
            if obj.get("kind", "").lower().strip() != kind:
                continue

            if not obj.get("metadata", {}).get("name"):
                if not skip_errors:
                    raise S2I2ThothException(f"No name provided in object found in {path!r}: %s", content)

                _LOGGER.error(
                    "No name provided in object found in %r: %s, skipping...", path, content,
                )
                continue

            if obj["metadata"]["name"] in result:
                if not skip_errors:
                    raise S2I2ThothException(
                        f"Duplicate name {path!r}, name already seen: {content}, seen "
                        f"in {result[obj['metadata']['name']]}"
                    )

                _LOGGER.error(
                    "Duplicate name %r, name already seen: %s, seen in %s, skipping...",
                    path,
                    content,
                    result[obj["metadata"]["name"]],
                )
                continue

            _LOGGER.info("Found %r of kind %r in %r", obj["metadata"]["name"], kind, path)
            result[obj["metadata"]["name"]] = class_(file_path=path, raw=obj)

        return result

    def to_yaml(self) -> str:
        """Convert the current object to its YAML representation."""
        result: str = yaml.safe_dump(self.raw)
        return result

    def save2file(self) -> None:
        """Save changes made to the file."""
        # This will work if the original file was is/bc. This will not work for templates
        # or list type - we need to find the given object and replace it correctly.
        with open(self.file_path, "w") as output_file:
            yaml.safe_dump(self.raw, output_file)

    def apply(self) -> None:
        """Apply changes to the cluster."""
        _LOGGER.info("Applying changes made to %r to the cluster", self.name)
        subcommand = _subprocess_run(
            ["oc", "apply", "-f", "-"], input=self.to_yaml(), stdout=subprocess.PIPE, stderr=subprocess.PIPE,
        )

        if subcommand.returncode != 0:
            raise OCError(
                f"Failed to apply {self.name!r} of kind {self.__class__.__name__!r}: {str(subcommand.stderr)}"
            )


@attr.s(slots=True)
class ImageStream(OpenShiftObject):
    """A wrapper for ImageStream representation."""

    @classmethod
    def load_file(cls, path: str, skip_errors: bool = True) -> Dict[str, "ImageStream"]:
        """Load ImageStream from a file."""
        content = cls._load_file_content(path, skip_errors=skip_errors)
        if content is None:
            return {}

        result: Dict[str, ImageStream] = cls._instantiate_objects(  # type: ignore
            path, content, cls, "imagestream", skip_errors=skip_errors
        )
        return result


@attr.s(slots=True)
class BuildConfig(OpenShiftObject):
    """A wrapper for BuildConfig representation."""

    _THOTH_ENV_VARS_DEFAULTS = {
        "UPGRADE_PIP_TO_LATEST": "",
        "DISABLE_SETUP_PY_PROCESSING": "",
        "THOTH_DRY_RUN": "0",
        "THOTH_ADVISE": "1",
        "THOTH_ASSEMBLE_DEBUG": "1",
        "THOTH_FROM_MASTER": "0",
        "THOTH_ERROR_FALLBACK": "1",
        "THAMOS_VERBOSE": "0",
        "THAMOS_FORCE": "0",
        "THAMOS_DEBUG": "0",
        "THAMOS_CONFIG_EXPAND_ENV": "0",
        "THAMOS_NO_PROGRESSBAR": "1",
        "THAMOS_NO_INTERACTIVE": "1",
        "THAMOS_REQUIREMENTS_FORMAT": "pipenv",
        "THAMOS_DEV": "0",
    }

    @classmethod
    def load_file(cls, path: str, skip_errors: bool = True) -> Dict[str, "BuildConfig"]:
        """Load BuildConfig from a file."""
        content = cls._load_file_content(path, skip_errors=skip_errors)
        if content is None:
            return {}

        result: Dict[str, BuildConfig] = cls._instantiate_objects(  # type: ignore
            path, content, cls, "buildconfig", skip_errors=skip_errors
        )
        return result

    def get_source_strategy(self) -> Dict[str, Any]:
        """Get source strategy entry out of the image stream."""
        result: Dict[str, Any] = self.raw.get("spec", {}).get("strategy", {}).get("sourceStrategy", {})
        return result

    def get_image_stream_name(self) -> str:
        """Get name of image stream without tag."""
        is_name: str = self.get_source_strategy().get("from", {}).get("name")
        is_name, _ = is_name.rsplit(":", maxsplit=1)
        return is_name

    def get_image_stream_tag_name(self) -> str:
        """Get name of image stream tag."""
        is_name: str = self.get_source_strategy().get("from", {}).get("name")
        _, is_tag = is_name.rsplit(":", maxsplit=1)
        return is_tag

    def get_image_stream_tag(self) -> str:
        """Get name of image stream tag."""
        result: str = self.get_source_strategy().get("from", {}).get("name")
        return result

    def set_image_stream_tag(self, image_stream_tag: str) -> None:
        """Setter for image stream tag."""
        self.get_source_strategy()["from"]["name"] = image_stream_tag

    def replace_in_file(self, old: str, new: str) -> None:
        """Replace the given string in the file."""
        # Note, changes are not reflected to this class - changes are in-file only.
        with open(self.file_path, "r") as f:
            content = f.read()

        # Do replace to keep formatting.
        content = content.replace(old, new)

        with open(self.file_path, "w") as f:
            f.write(content)

    def _do_trigger_build(self) -> None:
        """Trigger the build."""
        _LOGGER.info("Triggering build for %r in namespace %r", self.name, self.raw["metadata"]["namespace"])
        subcommand = _subprocess_run(
            ["oc", "start-build", "-n", self.raw["metadata"]["namespace"], self.name],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )

        if subcommand.returncode != 0:
            raise OCError(
                f"Failed to start build {self.name!r} in "
                f"namespace {self.raw['metadata']['namespace']!r}: {str(subcommand.stderr)}"
            )

    def trigger_build(self, only_if_no_config_change: bool = False) -> bool:
        """Trigger the build."""
        if only_if_no_config_change:
            for trigger in self.raw.get("triggers", []):
                if trigger.get("type") is not None and trigger["type"].lower() == "configchange":
                    return False
            else:
                self._do_trigger_build()
                return True

        self._do_trigger_build()
        return True

    def insert_thoth_env_vars(self) -> None:
        """Insert Thoth and Thamos specific environment variables."""
        _LOGGER.info("Inserting Thoth and Thamos specific environment variables to %r", self.name)
        source_strategy = self.get_source_strategy()
        if source_strategy is None:
            raise S2I2ThothException(
                "No source strategy defined for {self.name!r}, cannot inject environment variables"
            )

        if "env" not in source_strategy:
            source_strategy["env"] = []

        supplied_env = {k: v for k, v in os.environ.items() if k.startswith(("THOTH_", "THAMOS_"))}
        env_configuration = dict(self._THOTH_ENV_VARS_DEFAULTS)
        env_configuration.update(supplied_env)
        for thoth_env_name, thoth_env_value in env_configuration.items():
            for env_entry in source_strategy["env"]:
                if env_entry["name"] == thoth_env_name:
                    break
            else:
                source_strategy["env"].append(
                    {"name": thoth_env_name, "value": thoth_env_value}
                )



#!/usr/bin/env python3
# thoth-s2i
# Copyright(C) 2020 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""A library and tooling around Thoth's s2i."""

__version__ = "0.1.2"
__author__ = "Fridolin Pokorny <fridolin.pokorny@gmail.com>"

from .exceptions import ImportImageError
from .exceptions import OCError
from .exceptions import S2I2ThothException
from .lib import get_thoth_s2i_images
from .lib import import_thoth_s2i_image
from .lib import oc_check
from .lib import oc_get_bc



#!/usr/bin/env python3
# thoth-s2i
# Copyright(C) 2020 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
# type: ignore

"""A command line interface to thoth-s2i."""

import sys
import click
import logging
import json
import yaml
import tempfile
import re
from typing import Optional
from typing import List

import daiquiri

from thoth.s2i import __version__
from thoth.s2i.lib import get_thoth_s2i_images
from thoth.s2i.lib import import_thoth_s2i_image
from thoth.s2i.lib import oc_get_bc
from thoth.s2i.lib import oc_get_is
from thoth.s2i.lib import get_build_usage_report
from thoth.s2i.lib import change_image_stream
from thoth.s2i.objs import BuildConfig
from thoth.s2i.objs import ImageStream

from termcolor import colored

daiquiri.setup()
_LOGGER = logging.getLogger("thoth.s2i")
_LOGGER.setLevel(logging.INFO)

_DEFAULT_S2I_THOTH = "quay.io/thoth-station/s2i-thoth-ubi8-py36"
_DEFAULT_NAMESPACE = None

_MARK_OK = colored("✔", "green")
_MARK_BAD = colored("✖", "red")
_MARK_ENTRY = "📝"
_MARK_ARROW = "🠒"


@click.group()
@click.option("--debug/--no-debug", default=False, help="Run in a debug mode.")
def cli(debug: bool) -> None:
    """Convert an OpenShift application to use Thoth."""
    if debug:
        _LOGGER.setLevel(logging.DEBUG)


@cli.command()
def version() -> None:
    """Print thoth-s2i version and exit."""
    click.echo(f"thoth-s2i: {__version__}")
    sys.exit(1)


@cli.command("import-image")
@click.argument("image_names", required=False, metavar="THOTH_S2I_IMAGE_NAME:TAG", nargs=-1)
@click.option(
    "--namespace",
    "-n",
    required=_DEFAULT_NAMESPACE is None,
    default=_DEFAULT_NAMESPACE,
    show_default=True,
    metavar="NAMESPACE",
    help="OpenShift namespace to import Thoth images to.",
)
@click.option(
    "--all-latest",
    "-a",
    "all_images",
    required=False,
    is_flag=True,
    show_default=True,
    help="Import all images instead of prompting.",
)
@click.option(
    "--check-s2i-thoth/--no-check-s2i-thoth",
    "check_s2i_thoth",
    required=False,
    show_default=True,
    is_flag=True,
    default=True,
    help="Check the given image for availability in Thoth's s2i registry.",
)
def import_image(image_names: List[str], namespace: str, all_images: bool, check_s2i_thoth: bool) -> None:
    """Import Thoth image into an OpenShift deployment."""
    if all_images and image_names:
        _LOGGER.error("Cannot import all images if one or multiple image names were explicitly provided")
        sys.exit(1)

    if not (all_images or image_names):
        _LOGGER.error(
            "No image to import provided, state the image name explicitly or use --all-latest "
            "for importing all Thoth images with latest tag"
        )
        sys.exit(1)

    if not (image_names or all_images):
        image_names = [click.prompt("Choose Thoth image to be imported", default=_DEFAULT_S2I_THOTH)]

    if image_names and check_s2i_thoth:
        thoth_images = get_thoth_s2i_images()
        error = False
        for image_name in image_names:
            parts = image_name.rsplit(":", maxsplit=1)
            if len(parts) == 2:
                # Discard the image tag.
                image_name = parts[0]

            if image_name not in thoth_images:
                _LOGGER.error(
                    "Image %r not found in Thoth's images, available are: %s", image_name, thoth_images,
                )
                error = True

        if error:
            sys.exit(1)

    if not image_names:
        image_names = get_thoth_s2i_images()

    for image_name in image_names:
        _LOGGER.info("Importing image %r to namespace %r", image_name, namespace)
        import_thoth_s2i_image(namespace, image_name)


@cli.command("images")
@click.option(
    "--output-format",
    "-o",
    required=False,
    type=click.Choice(["pretty", "json", "yaml", "text"]),
    default="pretty",
    show_default=True,
    help="Output format for displaying Thoth images.",
)
def images(output_format: str) -> None:
    """Show available Thoth images."""
    thoth_images = get_thoth_s2i_images()

    if output_format == "json":
        click.echo(json.dumps({"s2i_thoth": thoth_images}, indent=2))
    elif output_format == "yaml":
        click.echo(yaml.safe_dump({"s2i_thoth": thoth_images}))
    elif output_format == "pretty":
        for image in thoth_images:
            click.echo(f"{_MARK_OK} {image}")
    elif output_format == "text":
        click.echo("\n".join(thoth_images))
    else:
        raise NotImplementedError(f"Unknown output format {output_format}")


@cli.command("report")
@click.option(
    "--namespace",
    "-n",
    required=_DEFAULT_NAMESPACE is None,
    default=_DEFAULT_NAMESPACE,
    show_default=True,
    metavar="NAMESPACE",
    help="OpenShift namespace to import Thoth images to.",
)
@click.option(
    "--output-format",
    "-o",
    required=False,
    type=click.Choice(["pretty", "json", "yaml"]),
    default="pretty",
    show_default=True,
    help="Output format for displaying Thoth compatibility.",
)
@click.option(
    "--selector",
    "-l",
    type=str,
    required=False,
    default=None,
    metavar="LABEL=SELECTOR",
    show_default=True,
    help="Label selector that should be applied to filter out BuildConfigs that should be migrated to Thoth's s2i.",
)
def report(namespace: str, selector: str, output_format: str) -> None:
    """Check the given namespace for s2i Thoth compatibility."""
    with tempfile.NamedTemporaryFile() as temp_file:
        oc_get_bc(namespace=namespace, selector=selector, path=temp_file.name)
        build_configs = BuildConfig.load_all(temp_file.name, skip_errors=True)

        if not build_configs:
            _LOGGER.warning("No BuildConfig found in namespace %r", namespace)
            sys.exit(1)

        oc_get_is(namespace=namespace, selector=selector, path=temp_file.name)
        image_streams = ImageStream.load_all(temp_file.name, skip_errors=True)

    report_dict = get_build_usage_report(image_streams, build_configs)
    if output_format == "json":
        click.echo(json.dumps(report_dict, indent=2))
    elif output_format == "yaml":
        click.echo(yaml.safe_dump(report_dict))
    elif output_format == "pretty":
        for build_config_name, build_config_value in report_dict.items():
            click.echo(f"{_MARK_ENTRY} {build_config_name}")
            for key, value in build_config_value.items():
                if value is None:
                    continue
                click.echo(f"\t\t{_MARK_ARROW} {key}: {value!r}")
    else:
        raise NotImplementedError(f"Unknown report output format {output_format!r}")


@cli.command("patch")
@click.argument("path", metavar="PATH")
@click.option(
    "--s2i-thoth",
    type=str,
    required=False,
    default=_DEFAULT_S2I_THOTH,
    show_default=True,
    prompt="Use s2i Thoth image",
    metavar="S2I_THOTH_IMAGE_NAME",
    help="Thoth's s2i image to be used.",
)
@click.option(
    "--check-s2i-thoth/--no-check-s2i-thoth",
    "check_s2i_thoth",
    required=False,
    show_default=True,
    is_flag=True,
    default=True,
    help="Check the given image for availability in Thoth's s2i registry.",
)
@click.option(
    "--tag", "-t", "tag", required=True, show_default=True, default="latest", help="Image stream tag to be used.",
)
@click.option(
    "--insert-env-vars/--no-insert-env-vars",
    "-e",
    "insert_env_vars",
    required=False,
    show_default=True,
    is_flag=True,
    default=False,
    help="Insert Thoth and Thamos specific environment variables into adjusted BuildConfigs.",
)
@click.option(
    "--from-image-stream-tag",
    "-f",
    required=False,
    show_default=True,
    type=str,
    default=".*python.*",
    help="A regular expression describing image stream tag that should be "
    "substituted with Thoth s2i image (a full match is applied).",
)
def patch(
    path: str,
    s2i_thoth: str,
    tag: str,
    from_image_stream_tag: str,
    check_s2i_thoth: bool = True,
    insert_env_vars: bool = True,
) -> None:
    """Patch local templates to use Thoth.

    Adjust templates stored on the filesystem to use Thoth's s2i.
    """
    if check_s2i_thoth:
        thoth_images = get_thoth_s2i_images()
        if s2i_thoth not in thoth_images:
            _LOGGER.error(
                "Image %r not found in Thoth's images, available are: %r", s2i_thoth, thoth_images,
            )
            sys.exit(1)

    _LOGGER.info("Patching files in %r to use %r", path, s2i_thoth)

    build_configs = BuildConfig.load_all(path, skip_errors=True)
    if not build_configs:
        _LOGGER.error("No BuildConfig found in %r", path)
        sys.exit(1)

    from_image_stream_tag_re = re.compile(from_image_stream_tag)

    image_stream_name = s2i_thoth.rsplit("/", maxsplit=1)[-1]
    changed_build_configs = change_image_stream(
        build_configs,
        f"{image_stream_name}:{tag}",
        lambda bc: bool(from_image_stream_tag_re.fullmatch(bc.get_image_stream_tag())),
    )

    if insert_env_vars:
        _LOGGER.warning(
            "Formatting of templates might change when inserting Thoth and Thamos specific environment variables"
        )
        for build_config in changed_build_configs.values():
            build_config.insert_thoth_env_vars()
            build_config.save2file()

    _LOGGER.info("Patching done \\o/, total BuildConfigs patched: %d", len(changed_build_configs))

    if len(changed_build_configs) > 0:
        _LOGGER.warning(
            "Don't forget to create an image stream with image %r and tag %r", s2i_thoth, tag,
        )


@cli.command("migrate")
@click.option(
    "--namespace",
    "-n",
    type=str,
    required=_DEFAULT_NAMESPACE is None,
    default=_DEFAULT_NAMESPACE,
    show_default=True,
    metavar="NAMESPACE",
    help="OpenShift namespace in which BuilConfigs should be adjusted.",
)
@click.option(
    "--s2i-thoth",
    type=str,
    required=False,
    default=None,
    show_default=True,
    metavar="S2I_THOTH_IMAGE_NAME",
    help="Thoth's s2i image to be used.",
)
@click.option(
    "--selector",
    "-l",
    type=str,
    required=False,
    default=None,
    metavar="LABEL=SELECTOR",
    show_default=True,
    help="Label selector that should be applied to filter out BuildConfigs that should be migrated to Thoth's s2i.",
)
@click.option(
    "--dry-run",
    is_flag=True,
    required=False,
    default=False,
    show_default=True,
    help="Do not propagate changes back to the cluster, rather print them to standard output for a review.",
)
@click.option(
    "--trigger-build",
    is_flag=True,
    required=False,
    default=False,
    show_default=True,
    help="Trigger build if BuildConfig has no build trigger on config change.",
)
@click.option(
    "--check-s2i-thoth/--no-check-s2i-thoth",
    "check_s2i_thoth",
    required=False,
    show_default=True,
    is_flag=True,
    default=True,
    help="Check the given image for availability in Thoth's s2i registry.",
)
@click.option(
    "--tag", "-t", "tag", required=True, show_default=True, default="latest", help="Image stream tag to be used.",
)
@click.option(
    "--import-image/--no-import-image",
    "do_import_image",
    required=False,
    show_default=True,
    is_flag=True,
    default=True,
    help="Image stream tag to be used.",
)
@click.option(
    "--insert-env-vars/--no-insert-env-vars",
    "-e",
    "insert_env_vars",
    required=False,
    show_default=True,
    is_flag=True,
    default=False,
    help="Insert Thoth and Thamos specific environment variables into adjusted BuildConfigs, "
    "environment variables can be also picked from the process environment - they start with "
    "THAMOS_ or THOTH_ prefix.",
)
@click.option(
    "--from-image-stream-tag",
    "-f",
    required=False,
    show_default=True,
    type=str,
    default=".*python.*",
    help="A regular expression describing image stream tag that should be "
    "substituted with Thoth s2i image (a full match is applied).",
)
def migrate(
    namespace: str,
    s2i_thoth: str,
    tag: str,
    from_image_stream_tag: str,
    selector: Optional[str] = None,
    do_import_image: bool = True,
    dry_run: bool = False,
    trigger_build: bool = False,
    check_s2i_thoth: bool = True,
    insert_env_vars: bool = True,
) -> None:
    """Migrate an existing OpenShift application to use Thoth.

    Adjust an existing OpenShift deployment to use Thoth by adjusting OpenShift's BuildConfigs.
    """
    s2i_thoth = s2i_thoth or click.prompt("Choose Thoth image to be imported", default=_DEFAULT_S2I_THOTH)

    if check_s2i_thoth:
        thoth_images = get_thoth_s2i_images()
        if s2i_thoth not in thoth_images:
            _LOGGER.error(
                "Image %r not found in Thoth's images, available are: %r", s2i_thoth, thoth_images,
            )
            sys.exit(1)

    from_image_stream_tag_re = re.compile(from_image_stream_tag)

    image_stream_name = s2i_thoth.rsplit("/", maxsplit=1)[-1]
    with tempfile.NamedTemporaryFile() as temp_file:
        oc_get_bc(namespace=namespace, selector=selector, path=temp_file.name)
        build_configs = BuildConfig.load_all(path=temp_file.name, skip_errors=True)

        changed_build_configs = change_image_stream(
            build_configs,
            f"{image_stream_name}:{tag}",
            lambda bc: bool(from_image_stream_tag_re.fullmatch(bc.get_image_stream_tag())),
        )

        if insert_env_vars:
            for build_config in changed_build_configs.values():
                build_config.insert_thoth_env_vars()

        if dry_run:
            if trigger_build:
                _LOGGER.warning("Dry run will not trigger build")

            if do_import_image:
                _LOGGER.warning("Dry run will not import image into OpenShift's registry")

            for build_config in changed_build_configs.values():
                click.echo("--")
                click.echo(build_config.to_yaml())
            return

        for build_config in changed_build_configs.values():
            build_config.apply()

        if len(changed_build_configs) > 0 and do_import_image:
            import_thoth_s2i_image(namespace, s2i_thoth)

        if trigger_build:
            for build_config in changed_build_configs.values():
                build_config.trigger_build(only_if_no_config_change=True)


__name__ == "__main__" and cli()
